'''This Python code is an automatically generated wrapper
for Fortran code made by 'fmodpy'. The original documentation
for the Fortran source code follows.

! Test Fortran COMPLEX wrapping and usage from Python with fmodpy.
'''

import os
import ctypes
import numpy

# --------------------------------------------------------------------
#               CONFIGURATION
# 
_verbose = True
_fort_compiler = "gfortran"
_shared_object_name = "test_complex256.so"
_this_directory = os.path.dirname(os.path.abspath(__file__))
_path_to_lib = os.path.join(_this_directory, _shared_object_name)
_compile_options = ['-fPIC', '-shared', '-O3']
_ordered_dependencies = ['test_complex256.f03', 'test_complex256_c_wrapper.f90']
# 
# --------------------------------------------------------------------
#               AUTO-COMPILING
#
# Try to import the existing object. If that fails, recompile and then try.
try:
    clib = ctypes.CDLL(_path_to_lib)
except:
    # Remove the shared object if it exists, because it is faulty.
    if os.path.exists(_shared_object_name):
        os.remove(_shared_object_name)
    # Compile a new shared object.
    _command = " ".join([_fort_compiler] + _compile_options + ["-o", _shared_object_name] + _ordered_dependencies)
    if _verbose:
        print("Running system command with arguments")
        print("  ", _command)
    # Run the compilation command.
    import subprocess
    subprocess.run(_command, shell=True, cwd=_this_directory)
    # Import the shared object file as a C library with ctypes.
    clib = ctypes.CDLL(_path_to_lib)
# --------------------------------------------------------------------


# This defines a C structure that can be used to hold the complex C type.
class c_complex_double(ctypes.Structure):
    _fields_ = [("real", ctypes.c_longdouble), ("imag", ctypes.c_longdouble)]
    # Define an "__init__" that can take a complex number as input.
    def __init__(self, real=0.0, imag=None):
        # If the provided first value was a complex number, expand it out.
        if (hasattr(real, "imag") and hasattr(real, "real") and (imag is None)):
            imag = real.imag
            real = real.real
        # Assign the internal fields.
        self.real = real
        self.imag = imag
    # Define a "value" so that it can be cleanly retreived as a single complex number.
    @property
    def value(self):
        return complex(self.real, self.imag)



# ----------------------------------------------
# Wrapper for the Fortran subroutine TEST_STANDARD

def test_standard(sing_in, array_in, array_out, known_array_out=None, known_matrix_out=None, opt_sing_in=None, opt_sing_out=None):
    '''! Test the basic functionaly of the 'COMPLEX' type and its
! interoperability with Python. This includes, inputs, outputs,
! array inputs with known and unknown size, optional inputs, and
! optional outputs.'''
    
    # Setting up "sing_in"
    if (type(sing_in) is not c_complex_double): sing_in = c_complex_double(sing_in)
    
    # Setting up "sing_out"
    sing_out = c_complex_double()
    
    # Setting up "array_in"
    if ((not issubclass(type(array_in), numpy.ndarray)) or
        (not numpy.asarray(array_in).flags.f_contiguous) or
        (not (array_in.dtype == numpy.dtype(numpy.complex128)))):
        import warnings
        warnings.warn("The provided argument 'array_in' was not an f_contiguous NumPy array of type 'numpy.complex128' (or equivalent). Automatically converting (probably creating a full copy).")
        array_in = numpy.asarray(array_in, dtype=numpy.complex128, order='F')
    array_in_dim_1 = ctypes.c_int(array_in.shape[0])
    
    # Setting up "array_out"
    if ((not issubclass(type(array_out), numpy.ndarray)) or
        (not numpy.asarray(array_out).flags.f_contiguous) or
        (not (array_out.dtype == numpy.dtype(numpy.complex128)))):
        import warnings
        warnings.warn("The provided argument 'array_out' was not an f_contiguous NumPy array of type 'numpy.complex128' (or equivalent). Automatically converting (probably creating a full copy).")
        array_out = numpy.asarray(array_out, dtype=numpy.complex128, order='F')
    array_out_dim_1 = ctypes.c_int(array_out.shape[0])
    
    # Setting up "known_array_out"
    if (known_array_out is None):
        known_array_out = numpy.zeros(shape=(array_out.size), dtype=numpy.complex128, order='F')
    elif ((not issubclass(type(known_array_out), numpy.ndarray)) or
          (not numpy.asarray(known_array_out).flags.f_contiguous) or
          (not (known_array_out.dtype == numpy.dtype(numpy.complex128)))):
        import warnings
        warnings.warn("The provided argument 'known_array_out' was not an f_contiguous NumPy array of type 'numpy.complex128' (or equivalent). Automatically converting (probably creating a full copy).")
        known_array_out = numpy.asarray(known_array_out, dtype=numpy.complex128, order='F')
    known_array_out_dim_1 = ctypes.c_int(known_array_out.shape[0])
    
    # Setting up "known_matrix_out"
    if (known_matrix_out is None):
        known_matrix_out = numpy.zeros(shape=(3, array_out.size), dtype=numpy.complex128, order='F')
    elif ((not issubclass(type(known_matrix_out), numpy.ndarray)) or
          (not numpy.asarray(known_matrix_out).flags.f_contiguous) or
          (not (known_matrix_out.dtype == numpy.dtype(numpy.complex128)))):
        import warnings
        warnings.warn("The provided argument 'known_matrix_out' was not an f_contiguous NumPy array of type 'numpy.complex128' (or equivalent). Automatically converting (probably creating a full copy).")
        known_matrix_out = numpy.asarray(known_matrix_out, dtype=numpy.complex128, order='F')
    known_matrix_out_dim_1 = ctypes.c_int(known_matrix_out.shape[0])
    known_matrix_out_dim_2 = ctypes.c_int(known_matrix_out.shape[1])
    
    # Setting up "opt_sing_in"
    opt_sing_in_present = ctypes.c_bool(True)
    if (opt_sing_in is None):
        opt_sing_in_present = ctypes.c_bool(False)
        opt_sing_in = c_complex_double()
    if (type(opt_sing_in) is not c_complex_double): opt_sing_in = c_complex_double(opt_sing_in)
    
    # Setting up "opt_sing_out"
    opt_sing_out_present = ctypes.c_bool(True)
    if (opt_sing_out is None):
        opt_sing_out_present = ctypes.c_bool(False)
        opt_sing_out = c_complex_double()

    print()
    print("sing_in ", sing_in)
    print("sing_out", sing_out)
    print("array_in ", array_in)
    print("array_out", array_out)

    # Call C-accessible Fortran wrapper.
    clib.c_test_standard(ctypes.byref(sing_in), ctypes.byref(sing_out), ctypes.byref(array_in_dim_1), ctypes.c_void_p(array_in.ctypes.data), ctypes.byref(array_out_dim_1), ctypes.c_void_p(array_out.ctypes.data), ctypes.byref(known_array_out_dim_1), ctypes.c_void_p(known_array_out.ctypes.data), ctypes.byref(known_matrix_out_dim_1), ctypes.byref(known_matrix_out_dim_2), ctypes.c_void_p(known_matrix_out.ctypes.data), ctypes.byref(opt_sing_in_present), ctypes.byref(opt_sing_in), ctypes.byref(opt_sing_out_present), ctypes.byref(opt_sing_out))

    print()
    print("sing_in ", sing_in)
    print("sing_out", sing_out)
    print("array_in ", array_in)
    print("array_out", array_out)

    # Return final results, 'INTENT(OUT)' arguments only.
    return sing_out.value, array_out, known_array_out, known_matrix_out, (opt_sing_out.value if opt_sing_out_present else None)


# ----------------------------------------------
# Wrapper for the Fortran subroutine TEST_EXTENDED

def test_extended(n, opt_array_in=None, known_opt_array_out=None, opt_alloc_array_out=None):
    '''! Test the extended functionaly of the 'COMPLEX' type and its
! interoperability with Python. This includes, optional array
! inputs, optional array outputs, and allocatable array outputs.'''
    
    # Setting up "opt_array_in"
    opt_array_in_present = ctypes.c_bool(True)
    if (opt_array_in is None):
        opt_array_in_present = ctypes.c_bool(False)
        opt_array_in = numpy.zeros(shape=(1), dtype=numpy.complex128, order='F')
    elif (type(opt_array_in) == bool) and (opt_array_in):
        opt_array_in = numpy.zeros(shape=(1), dtype=numpy.complex128, order='F')
    elif ((not issubclass(type(opt_array_in), numpy.ndarray)) or
          (not numpy.asarray(opt_array_in).flags.f_contiguous) or
          (not (opt_array_in.dtype == numpy.dtype(numpy.complex128)))):
        import warnings
        warnings.warn("The provided argument 'opt_array_in' was not an f_contiguous NumPy array of type 'numpy.complex128' (or equivalent). Automatically converting (probably creating a full copy).")
        opt_array_in = numpy.asarray(opt_array_in, dtype=numpy.complex128, order='F')
    if (opt_array_in_present):
        opt_array_in_dim_1 = ctypes.c_int(opt_array_in.shape[0])
    else:
        opt_array_in_dim_1 = ctypes.c_int()
    
    # Setting up "known_opt_array_out"
    known_opt_array_out_present = ctypes.c_bool(True)
    if (known_opt_array_out is None):
        known_opt_array_out_present = ctypes.c_bool(False)
        known_opt_array_out = numpy.zeros(shape=(1), dtype=numpy.complex128, order='F')
    elif (type(known_opt_array_out) == bool) and (known_opt_array_out):
        known_opt_array_out = numpy.zeros(shape=(3), dtype=numpy.complex128, order='F')
    elif ((not issubclass(type(known_opt_array_out), numpy.ndarray)) or
          (not numpy.asarray(known_opt_array_out).flags.f_contiguous) or
          (not (known_opt_array_out.dtype == numpy.dtype(numpy.complex128)))):
        import warnings
        warnings.warn("The provided argument 'known_opt_array_out' was not an f_contiguous NumPy array of type 'numpy.complex128' (or equivalent). Automatically converting (probably creating a full copy).")
        known_opt_array_out = numpy.asarray(known_opt_array_out, dtype=numpy.complex128, order='F')
    if (known_opt_array_out_present):
        known_opt_array_out_dim_1 = ctypes.c_int(known_opt_array_out.shape[0])
    else:
        known_opt_array_out_dim_1 = ctypes.c_int()
    
    # Setting up "opt_alloc_array_out"
    opt_alloc_array_out_present = ctypes.c_bool(True)
    if (opt_alloc_array_out is None):
        opt_alloc_array_out_present = ctypes.c_bool(False)
        opt_alloc_array_out = numpy.zeros(shape=(0), dtype=numpy.complex128, order='F')
    elif (type(opt_alloc_array_out) == bool) and (opt_alloc_array_out):
        opt_alloc_array_out = numpy.zeros(shape=(1), dtype=numpy.complex128, order='F')
    elif ((not issubclass(type(opt_alloc_array_out), numpy.ndarray)) or
          (not numpy.asarray(opt_alloc_array_out).flags.f_contiguous) or
          (not (opt_alloc_array_out.dtype == numpy.dtype(numpy.complex128)))):
        import warnings
        warnings.warn("The provided argument 'opt_alloc_array_out' was not an f_contiguous NumPy array of type 'numpy.complex128' (or equivalent). Automatically converting (probably creating a full copy).")
        opt_alloc_array_out = numpy.asarray(opt_alloc_array_out, dtype=numpy.complex128, order='F')
    if (opt_alloc_array_out_present):
        opt_alloc_array_out_dim_1 = ctypes.c_int(opt_alloc_array_out.shape[0])
    else:
        opt_alloc_array_out_dim_1 = ctypes.c_int()
    opt_alloc_array_out = ctypes.c_void_p(opt_alloc_array_out.ctypes.data)
    
    # Setting up "n"
    if (type(n) is not ctypes.c_int): n = ctypes.c_int(n)
    
    # Setting up "alloc_array_out"
    alloc_array_out = ctypes.c_void_p()
    alloc_array_out_dim_1 = ctypes.c_int()

    # Call C-accessible Fortran wrapper.
    clib.c_test_extended(ctypes.byref(opt_array_in_present), ctypes.byref(opt_array_in_dim_1), ctypes.c_void_p(opt_array_in.ctypes.data), ctypes.byref(known_opt_array_out_present), ctypes.byref(known_opt_array_out_dim_1), ctypes.c_void_p(known_opt_array_out.ctypes.data), ctypes.byref(opt_alloc_array_out_present), ctypes.byref(opt_alloc_array_out_dim_1), ctypes.byref(opt_alloc_array_out), ctypes.byref(n), ctypes.byref(alloc_array_out_dim_1), ctypes.byref(alloc_array_out))

    # Post-processing "opt_alloc_array_out"
    opt_alloc_array_out_size = (opt_alloc_array_out_dim_1.value)
    if (opt_alloc_array_out_present) and (opt_alloc_array_out_size > 0):
        opt_alloc_array_out = numpy.array(ctypes.cast(opt_alloc_array_out, ctypes.POINTER(c_complex_double*opt_alloc_array_out_size)).contents, copy=False).view(numpy.complex128)
    elif (opt_alloc_array_out_size == 0):
        opt_alloc_array_out = numpy.zeros(opt_alloc_array_out_dim_1.value, dtype=numpy.complex128, order='F')
    else:
        opt_alloc_array_out = None
    
    # Post-processing "alloc_array_out"
    alloc_array_out_size = (alloc_array_out_dim_1.value)
    if (alloc_array_out_size > 0):
        alloc_array_out = numpy.array(ctypes.cast(alloc_array_out, ctypes.POINTER(c_complex_double*alloc_array_out_size)).contents, copy=False).view(numpy.complex128)
    elif (alloc_array_out_size == 0):
        alloc_array_out = numpy.zeros(alloc_array_out_dim_1.value, dtype=numpy.complex128, order='F')
    else:
        alloc_array_out = None
    
    # Return final results, 'INTENT(OUT)' arguments only.
    return (known_opt_array_out if known_opt_array_out_present else None), (opt_alloc_array_out if opt_alloc_array_out_present else None), alloc_array_out






# Standardized testing interface.
def test():
    import os
    dir_name = os.path.dirname(os.path.abspath(__file__))
    test_name = os.path.basename(dir_name)
    fort_file = os.path.join(dir_name, f"{test_name}.f03")
    build_dir = os.path.join(dir_name, f"fmodpy_{test_name}")
    print(f"  {test_name}..", end=" ", flush=True)
    import fmodpy
    fort = fmodpy.fimport(fort_file, build_dir=build_dir,
                          output_dir=dir_name, verbose=False, wrap=True,
                          rebuild=True, show_warnings=False)
    # ---------------------------------------------------------------
    # Begin specific testing code.

    import numpy as np
    a = np.array([True, False, True, False, True, False, True, False], dtype=bool)
    temp = np.asarray(a, dtype='int32')
    out = temp.copy()
    assert(all(out == fort.test_simple_logical(temp, b=out, c=False)))
    assert(not any(fort.test_simple_logical(temp, b=out, c=True)))

    # End specific testing code.
    # ---------------------------------------------------------------
    print("passed", flush=True)


if __name__ == "__main__":
    test()




# Standardized testing interface.
def test():
    import os
    dir_name = os.path.dirname(os.path.abspath(__file__))
    test_name = os.path.basename(dir_name)
    fort_file = os.path.join(dir_name, f"test_{test_name}.f03")
    build_dir = os.path.join(dir_name, f"fmodpy_{test_name}")
    print(f"  {test_name}..", end=" ", flush=True)
    import fmodpy
    fort = fmodpy.fimport(fort_file, build_dir=build_dir,
                          output_dir=dir_name, rebuild=True,
    )
    # ---------------------------------------------------------------
    # Begin specific testing code.

    import numpy as np
    n = 10
    array_test = np.asarray(np.arange(n), dtype="float32", order='F')

    # Assign some internal variables.
    #   a_pub (float)
    fort.addition.a_pub = 1.5
    assert(fort.addition.a_pub == 1.5)
    fort.addition.a_pub = 2.5
    assert(fort.addition.a_pub == 2.5)

    #   b_pub (integer)
    fort.addition.b_pub = 2
    assert(fort.addition.b_pub == 2)
    fort.addition.b_pub = 1
    assert(fort.addition.b_pub == 1)

    # Verify that the public subtraction routine works correctly.
    assert(3.5 == fort.addition.add_ab())

    #   a_vec_pub (real vector of size 10)
    import numpy as np
    a = np.asarray(np.random.random(size=(10,)), dtype=np.float32)
    # testing using the "get" before assignment
    fort.addition.a_vec_pub 
    # test assignment.
    fort.addition.a_vec_pub = a
    assert(all(a == fort.addition.a_vec_pub))
    # try assigning with a vector that is too small..
    a = np.asarray(np.random.random(size=(8,)), dtype=np.float32)
    fort.addition.a_vec_pub = a
    assert(all(a == fort.addition.a_vec_pub[:len(a)]))
    # try assigning with a vector that is too big, might seg-fault..
    a = np.asarray(np.random.random(size=(80,)), dtype=np.float32)
    fort.addition.a_vec_pub = a
    assert(all(a[:len(fort.addition.a_vec_pub)] == fort.addition.a_vec_pub))
    # reassign a normal value.
    a = np.asarray(np.random.random(size=(10,)), dtype=np.float32)
    fort.addition.a_vec_pub = a

    # Make sure the vector is correctly negated internally and returned.
    result = fort.addition.sub_vecs()
    assert(all(result == -a))

    #   b_vec_pub (real vector of size 10)
    b = np.asarray(np.arange(4), dtype=np.float32)
    # testing using the "get" before assignment
    # test assignment.
    fort.addition.b_vec_pub = b
    assert(all(b == fort.addition.b_vec_pub))
    # try assigning over top with a new vector
    b = np.asarray(np.arange(10), dtype=np.float32)
    fort.addition.b_vec_pub = b
    assert(all(b == fort.addition.b_vec_pub))

    # Make sure the vectors are correctly subtracted internally and returned.
    result = fort.addition.sub_vecs()
    assert(all(b == result + a))

    # TODO: check size of array, if large enough, warn about copy in SETTER
    # TODO: provide some protection against incorrectly sized array assignments

    # End specific testing code.
    # ---------------------------------------------------------------
    print("passed", flush=True)
    import shutil
    shutil.rmtree(os.path.join(dir_name,f"test_{test_name}"))

if __name__ == "__main__":
    test()



import os, sysconfig
import numpy as NP

# Default configurable variables.
omp                = False
blas               = False
lapack             = False
verbose            = False
verbose_module     = True
autocompile        = True
wrap               = True
rebuild            = False
show_warnings      = True
debug_line_numbers = False
implicit_typing    = False
end_is_named       = True
overwrite          = False # hard-delete destination directory if it already exists
log_file           = os.devnull
f_compiler         = "gfortran"
f_compiler_args    = ["-fPIC", "-shared", "-O3"]
link_omp           = ["-fopenmp"]
link_blas          = ["-lblas"]
link_lapack        = ["-lblas", "-llapack"]
home_directory     = os.path.expanduser("~")
libraries          = NP.__path__ + [
    "/usr/lib",
    "/lib",
    "/usr/lib64",
    "/lib64",
    "/usr/lib32",
    "/lib32",
    "/opt/homebrew/Cellar/openblas",
    "/opt/homebrew/Cellar/libomp"
]
library_recursion  = 2
library_extensions = ["so", "dylib"]
symbol_command     = 'nm -gU "{path}" 2> /dev/null || nm -gD "{path}" 2> /dev/null'
config_file        = ".fmodpy.py"
wait_warning_sec   = 5 # number of seconds to wait before warning about automatic compilation

# --------------------------------------------------------------------
#      Development globals, not intended to be changed by users.
# 
# All of these variables should have expected types.
BOOL_CONFIG_VARS = ["omp", "blas", "lapack", "verbose", "autocompile",
                    "wrap", "rebuild", "show_warnings",
                    "debug_line_numbers", "implicit_typing",
                    "end_is_named", "overwrite"]
LIST_CONFIG_VARS = ["f_compiler_args", "link_omp", "link_blas",
                    "link_lapack", "libraries", "library_extensions"]
# File related maniplation arguments
PY_EXT = ".py"
FORT_EXT = ".f90"
PYTHON_WRAPPER_EXT = "_python_wrapper"
FORT_WRAPPER_EXT = "_c_wrapper"+FORT_EXT
GET_SIZE_PROG_FILE = "fmodpy_get_size"+FORT_EXT
GET_SIZE_EXEC_FILE = "fmodpy_get_size"
# --------------------------------------------------------------------

# Automatically handle printing for status updates.
#   WARNING: Do not use this function for warnings. Use `warnings.warn`.
#   WARNING: Do not use this function for errors. Use `raise(...)`.
# 
# Custom print function (allows for line numbers to be automatically
# added to all print statements, easily controls verbosity level).
def fmodpy_print(*args, **kwargs):
    # Skip all print statements if verbosity is off.
    global verbose
    if (not verbose): return
    # Set the log file.
    global log_file
    if (log_file == os.devnull): 
        import sys
        log_file = sys.stdout
    # Add information about where the bit is printed from, if turned on.
    global debug_line_numbers    
    if debug_line_numbers:
        import inspect
        # Add the calling file and line to the print statement
        calling_function = inspect.stack()[1]
        calling_file = calling_function.filename
        calling_file = os.path.basename(calling_file)
        calling_line = calling_function.lineno
        args += (f'({calling_file} line {calling_line})',) 
    # For all print operations, force a "flush" output.
    kwargs["file"] = log_file
    kwargs["flush"] = True
    print(*args, **kwargs)


# Execute a blocking command with a subprocess, on completion provide
# the return code, stdout as string, and stderr as string. This should
# work across all Python3.x as well as cross-platform.
#  INPUT:
#   command -- A list of strings or string (space separated) describing
#              a standard command as would be given to subprocess.Popen
# 
#  OUTPUT:
#   return_code -- Straight from the subprocess returncode
#   stdout      -- A list of strings that are the lines of stdout 
#                  produced by the execution of <command>
#   stderr      -- A list of strings that are the lines of stderr
#                  produced by the execution of <command>
def run(command, **popen_kwargs):
    import sys, subprocess
    # For Python 3.x and x < 6, the encoding is a string by default
    # For Python 3.6 and later the encoding can be given as an arguemnt
    if sys.version_info >= (3,6):
        popen_kwargs.update( dict(encoding="UTF-8") )
    proc = subprocess.Popen(command, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, **popen_kwargs)
    stdout, stderr = proc.communicate()
    # Before python 3.6, the encoding had to be handled after collection
    if (sys.version_info < (3,6)):
        if (type(stdout) != str): stdout = str(stdout, encoding="UTF-8")
        if (type(stderr) != str): stderr = str(stderr, encoding="UTF-8")
    # Remove Windows specific characters and split by the new line.
    if stdout: stdout = stdout.replace("\r","").split("\n")
    else:      stdout = ""
    if stderr: stderr = stderr.replace("\r","").split("\n")
    else:      stderr = ""
    # Return the exit code, standard out, and standard error.
    return proc.returncode, stdout, stderr


# Configure the current 'fmodpy' runtime. Start with default values in
# this file, load any global defaults over those, and finally override
# again with provided keyword arguments.
def load_config(**kwargs):
    # Get the "globals" for this file, this is the configuration of fmodpy.
    fmodpy_config = globals()

    # Identify those elements of "fmodpy.config" that should not be
    # set or considered when printing out configuration.
    modules = {"os", "sysconfig"}
    func_type = type(lambda:None)
    functions = {k for k in fmodpy_config if
                 type(fmodpy_config[k]) == func_type}
    # Set the default configuration as the current configuration.
    config = { k:fmodpy_config[k] for k in fmodpy_config
               if (k[0].isalpha() and k.islower()) and
               (k not in modules) and (k not in functions) }

    # Check to make sure that all variables given specify only those
    # that are allowed to be specified.
    for var in kwargs:
        if (var not in config):
            from fmodpy.exceptions import UnrecognizedConfiguration
            raise(UnrecognizedConfiguration(f"Configuration for '{var}' is given, but that variable does not exist."))
        if (var[0].isupper() or (not var[0].isalpha())):
            from fmodpy.exceptions import IllegalConfiguration
            raise(IllegalConfiguration(f"The variable '{var}' is not allowed to be configured."))

    # Update the 'config' dictionary with the provided keyword arguments.
    config.update( kwargs )

    # Make sure the path names do not have spaces.
    if any(' ' in str(config[k]) for k in ('f_compiler',)):
        from fmodpy.exceptions import NotAllowedPath
        raise(NotAllowedPath("Spaces cannot be included in compiler or linker paths."))

    # Convert list-type configurations into lists (from strings).
    for var in LIST_CONFIG_VARS:
        if (type(config[var]) is str):
            config[var] = config[var].split()
        elif (type(config[var]) is not list):
            from fmodpy.exceptions import IllegalConfiguration
            raise(IllegalConfiguration(f"The variable '{var}' is supposed to be a list or string, but is neither."))

    # Make sure that boolean-type configurations are booleans (from strings).
    for var in BOOL_CONFIG_VARS:
        if (type(config[var]) is str):
            config[var] = (config[var].lower().strip() == 'true')
        elif (type(config[var]) is not bool):
            from fmodpy.exceptions import IllegalConfiguration
            raise(IllegalConfiguration(f"The variable '{var}' is supposed to be a bool or string, but is neither."))

    # If 'lblas' is True, then add BLAS compilation and
    # link arguments to the list of arguments already.
    if config["blas"]:
        for l in config["link_blas"]:
            if (l not in config["f_compiler_args"]):
                config["f_compiler_args"] += [l]

    # If 'llapack' is True, then add LAPACK compilation and
    # link arguments to the list of arguments already.
    if config["lapack"]:
        for l in config["link_lapack"]:
            if (l not in config["f_compiler_args"]):
                config["f_compiler_args"] += [l]

    # If 'omp' is True, then add OpenMP compilation and link arguments
    # to the list of arguments already.
    if config["omp"]:
        for l in config["link_omp"]:
            if (l not in config["f_compiler_args"]):
                config["f_compiler_args"] += [l]

    # Set all of the configuration variables as module-wide globals.
    for var in config: fmodpy_config[var] = config[var]

    # Return the current configuration.
    return config


# --------------------------------------------------------------------
# --------------------------------------------------------------------


# Read a configuration file into a Python dictionary.
def read_config_file(path):
    with open(path) as f:
        # Get all lines that have one "=" split and stripped into nice strings.
        # Strip quote characters from edges of strings (when user assigned a string).
        lines = ((' '.join(var.strip().strip('\'"').split())  for var in l.strip().split('='))
                 for l in f.readlines() if (l.count('=') == 1))
    return dict(lines)

# Load local settings if they are available.
if os.path.exists(os.path.join(home_directory,config_file)):
    global_config = read_config_file(
        os.path.join(home_directory,config_file))
else: global_config = {}
# Make sure all provided settings are recognized (and allowable).
for var in global_config:
    if (var.strip()[:1] == "#"): continue # Skip commented lines.
    if var not in globals():
        from fmodpy.exceptions import UnrecognizedConfiguration
        raise(UnrecognizedConfiguration(f"Configuration for '{var}' is given in config file, but that variable does not exist."))
# Update the configuration of fmodpy.
globals().update(global_config)
# Delete the local variables (so they don't hang around).
if (len(global_config) > 0): del var
del read_config_file
del global_config
# Load in the global configuration file.
load_config()




# Custom types:
#   - read in the "arguments" for the type, those are the sizes
#   - allow "LEN" arguments
#   - prohibit "KIND" arguments
#   - define a "struct generator" in python that creates a struct
#     of the same shape as in Fortran given the same arguments
#   - check for "issubclass" instead of exact type match for structs
#   



# --------------------------------------------------------------------
# 
#                         DEVELOPMENT
#                            PLAN
# 
# - support custom types defined in Fortran
# --- parse the custom type and determine its contents in terms of basic types
# --- generate a matching C struct using ctypes to the custom type
# --- identify proper allocation schema in numpy to make sure memory
#     maps correctly into Fortran
# 
# - support procedures as arguments
# --- create a registry of "fmodpy wrapped functions" with the memory
#     addresses of all functions wrapped by fmodpy
# --- map the memory address to the actual function in the Fortran
#     wrapper
# 
# --- construct C function in Python (with ctypes) that translates C 
#     style data into Python compatible data, calls the Python 
#     function that was passed in as an argument, then translates the
#     results back into C style data for return
# --- construct a BIND(C) Fortran wrapper function that translates
#     arguments from Fortran style into C style and calls the C 
#     (Python) function from Fortran, nested in a MODULE with PUBLIC
#     variables that assign the address of the C function
# --- pass the memory address of the C function in Python to Fortran
#     wrapper, use that address to assign a local Fortran function 
#     in the BIND(C) wrapper module that calls Python C function
# 
# - extend test cases to achieve (near) 100% coverage of all lines
# 
# 
# --------------------------------------------------------------------
#                            TODO LISTS
# 
# FMODPY
# - Create WARNING for when link files are "not found" in compiled
#   projects, asking if the links are still alive.
# - Make __init__.py the source file and the _python_wrapper.py the
#   link file, so that projects put on GitHub will work better.
# - Write a code that "recovers links" instituted by GitHub, insert
#   that into the generated fmodpy wrappers.
# - Create an archive of the source files when the wrapper was originally
#   built, so that those can be recovered even if the links are lost.
# - add option to continue wrapping on failures, and to simply skip
#   wrappers for the things that cannot be wrapped (when possible)
# - assign implicit_typing to True and named_end to False for '.f' files
# - assign correct shape (from shape[0] ...) for implicit array arguments
# - check the memory contiguity of OPTIONAL input arrays 
# - LOGICAL array warning should only happen *after* non-arguments are deleted
# - compile dependencies first when 'dependencies' are given and 'autocompile=True'
# - remove automatic size option in wrapper for optional arguments with unknown shape
# - handle automatic shapes that have a ":" inside of them
# - check for existence of Fortran compiler on install, offer instruction
# - support lists of compilation commands to run before making final object
# - support list of files to include in final module compilation (like objects)
# - support multiple file input, for specifying all files that are part
#   of a module or may contain important types or constants
# - empty classes should not be added into wrappers, they should be skipped
#   (i.e., a class that only defines interfaces)
# - automatic documentation of the types of inputs in Python, included
#   in addition to original Fortran documentation.
# - replace names of builtins like "global" and "lambda" in the
#   python wrapper (when the original Fortran uses these names).
# - IMPLICIT NONE should be added to interfaces for functions
# - integer arguments that are used as sizes should be declared before
#   the array arguments that use them as sizes, regardless of their
#   order in the actual list of arguments
# 
# TYPE
# - find the definition of this Type, and all of its contents.
# - the type could be declared in a module that is "USED" in
#   another file, will have to scan other files for modules
#   with type definitions?
#    in the same subroutine or function
#    in the same module
#    in a USED module, potentially in another file
# - define a C struct in the appropriate context 
# - cycle all arugments in type, initializing them correctly
# - store all the initialized values into a C struct
# 
# 
# PROCEDURE
# - The interface being used by a subroutine-as-an-argument could
#   be defined in another module. Need to check for that.
#   Need to disallow "EXTERNAL" subroutines-as-arguments.
# - Need to write fort_to_py versions of interface functions in
#   fortran wrapper and in python wrapper
# - Have a way to bypass Python when a fmodpy wrapped fortran routine
#   is given. Maybe define a get_address function for all source
#   functions in Fortran wrappers?
# 
